/*
 * This file is part of Remote X11 (RX11), a library and server that lets
 * you develop applications on systems without X11 support.
 *
 * Copyright (C) 1995, Kay Roemer. See COPYING for details.
 */

#ifndef _RXLIB_H
#define _RXLIB_H

#include <sys/types.h>
#include "X.h"

#define RXConnectionNumber(dpy)     ((dpy)->fd)
#define RXRootWindow(dpy, scr) 	    ((dpy)->screen->root)
#define RXDefaultScreen(dpy)        ((dpy)->server_screen)
#define RXDefaultRootWindow(dpy)    ((dpy)->screen->root)
#define RXDefaultVisual(dpy, scr)   ((dpy)->screen->root_visual)
#define RXDefaultGC(dpy, scr) 	    ((dpy)->screen->default_gc)
#define RXBlackPixel(dpy, scr) 	    ((dpy)->screen->black_pixel)
#define RXWhitePixel(dpy, scr) 	    ((dpy)->screen->white_pixel)
#define RXAllPlanes                 ((unsigned long)~0L)
#define RXDisplayWidth(dpy, scr)    ((dpy)->screen->width)
#define RXDisplayHeight(dpy, scr)   ((dpy)->screen->height)
#define RXDisplayWidthMM(dpy, scr)  ((dpy)->screen->mwidth)
#define RXDisplayHeightMM(dpy, scr) ((dpy)->screen->mheight)
#define RXDisplayPlanes(dpy, scr)   ((dpy)->screen->root_depth)
#define RXDisplayCells(dpy, scr)    (RXDefaultVisual((dpy), (scr))->map_entries)
#define RXScreenCount(dpy)          (1)
#define RXDefaultDepth(dpy, scr)    ((dpy)->screen->root_depth)
#define RXDefaultColormap(dpy, scr) ((dpy)->screen->cmap)


typedef GContext RXGC;

typedef struct {
  VisualID visualid;
  short    class;
  u_long   red_mask, green_mask, blue_mask;
  short    bits_per_rgb;
  short    map_entries;
} RXVisual;

typedef struct {
  Window   root;
  short    width, height;
  short    mwidth, mheight;
  short    root_depth;
  RXVisual *root_visual;
  RXGC     default_gc;
  Colormap cmap;
  u_long   white_pixel;
  u_long   black_pixel;
} RXScreen;

typedef struct {
  short    fd;
  RXScreen *screen;
  u_long   server_addr;
  short    server_screen;
  short    server_display;
} RXDisplay;

typedef struct {
  u_long mask;
  short  function;
  u_long plane_mask;
  u_long foreground;
  u_long background;
  short  line_width;
  short  line_style;
  short  cap_style;
  short  join_style;
  short  fill_style;
  short  arc_mode;
  Pixmap tile;
  Pixmap stipple;
  short  ts_x_origin;
  short  ts_y_origin;
  Font   font;
  short  subwindow_mode;
  short  clip_x_origin;
  short  clip_y_origin;
  Pixmap clip_mask;
  short  dash_offset;
  short  dashes;
} RXGCValues;

typedef struct {
  u_long   mask;
  Pixmap   background_pixmap;
  u_long   background_pixel;
  Pixmap   border_pixmap;
  u_long   border_pixel;
  Colormap colormap;
  Cursor   cursor;
} RXSetWindowAttributes;

typedef struct {
  short    x, y;
  short    width, height;
  short    border_width;
  short    depth;
  RXVisual *visual;
  short    class;
  Colormap colormap;
  short    colormap_installed;
  short    map_state;
} RXWindowAttributes;

typedef struct {
  u_long mask;
  short  x, y;
  short  width, height;
  short  border_width;
  Window sibling;
  short  stack_mode;
} RXWindowChanges;

typedef struct {
  u_long  pixel;
  u_short red, green, blue;
  short   flags;
} RXColor;

typedef struct {
  short x1, y1, x2, y2;
} RXSegment;

typedef struct {
  short x, y;
} RXPoint;

typedef struct {
  short   x, y;
  u_short width, height;
} RXRectangle;

typedef struct {
  short   x, y;
  u_short width, height;
  short   angle1, angle2;
} RXArc;

typedef struct {
  short   lbearing;
  short   rbearing;
  short   width;
  short   ascent;
  short   descent;
  u_short attributes;
} RXCharStruct;

typedef struct {
  Font           fid;
  u_short        direction;
  u_short        min_char_or_byte2;
  u_short        max_char_or_byte2;
  u_short        min_byte1;
  u_short        max_byte1;
  short          all_chars_exist;
  u_short        default_char;
  RXCharStruct   min_bounds;
  RXCharStruct   max_bounds;
  int            ascent;
  int            descent;
} RXFontStruct;


extern void       RXFree (void *);

extern RXDisplay *RXOpenDisplay (char *);
extern int        RXCloseDisplay (RXDisplay *);


extern RXGC       RXCreateGC (RXDisplay *, Drawable, u_long m, RXGCValues *);
extern void       RXChangeGC (RXDisplay *, RXGC, u_long mask, RXGCValues *);
extern void       RXFreeGC   (RXDisplay *, RXGC);


extern Window     RXCreateSimpleWindow (RXDisplay *, Window parent,
					int x, int y, u_int w, u_int h,
					u_int border_w,
					u_long border, u_long bg);
extern void       RXDestroyWindow (RXDisplay *, Window);
extern void       RXStoreName (RXDisplay *, Window, char *);


extern void       RXDrawPoint (RXDisplay *, Drawable, RXGC, int x, int y);
extern void       RXDrawLine  (RXDisplay *, Drawable, RXGC,
			       int x1, int y1, int x2, int y2);
extern void       RXDrawRectangle (RXDisplay *, Drawable, RXGC, int x, int y,
				   u_int w, u_int h);
extern void       RXDrawArc (RXDisplay *, Drawable, RXGC, int x, int y,
			     u_int w, u_int h, short a1, short a2);
extern void       RXDrawString (RXDisplay *, Drawable, RXGC, int x, int y,
				char *, int len);
extern void       RXDrawImageString (RXDisplay *, Drawable, RXGC, int x, int y,
				     char *, int len);

extern void       RXFillRectangle (RXDisplay *, Drawable, RXGC, int x, int y,
				   u_int w, u_int h);
extern void       RXFillArc (RXDisplay *, Drawable, RXGC, int x, int y,
			     u_int w, u_int h, short a1, short a2);


extern Pixmap     RXCreatePixmap (RXDisplay *, Drawable, int w, int h, int d);
extern Pixmap     RXCreateBitmapFromData (RXDisplay *, Drawable, char *,
					  int w, int h);
extern Pixmap     RXCreatePixmapFromBitmapData (RXDisplay *, Drawable, char *,
						int w, int h,
						u_long fg, u_long bg,
						int d);
extern void       RXFreePixmap (RXDisplay *, Pixmap);

extern char      *RXReadXbmFile (RXDisplay *, char *fname, u_int *w, u_int *h,
				 int *xhot, int *yhot);
extern int        RXReadBitmapFile (RXDisplay *dpy, Drawable d, char *fname,
				    u_int *w, u_int *h, Pixmap *pix,
				    int *xhot, int *yhot);
extern char      *RXScaleBitmapData (RXDisplay *, u_char *data,
				     u_int *w, u_int *h,
				     u_int vscale, u_int hscale);


extern void       RXCopyArea (RXDisplay *, Drawable s, Drawable d, RXGC,
			      int sx, int sy, u_int w, u_int h,
			      int dx, int dy);
extern void       RXCopyPlane (RXDisplay *, Drawable s, Drawable d, RXGC,
			       int sx, int sy, u_int w, u_int h,
			       int dx, int dy, u_long p);
extern void       RXCopyBitmap (RXDisplay *, Drawable d, RXGC, char *data,
				u_int width, u_int height,
				int sx, int sy, u_int w, u_int h,
				int dx, int dy);


extern int        RXAllocColor (RXDisplay *, Colormap, RXColor *);
extern int        RXAllocNamedColor (RXDisplay *, Colormap, char *,
				     RXColor *scr_def, RXColor *exact_def);
extern int        RXQueryColor (RXDisplay *, Colormap, RXColor *);

extern void       RXFlush (RXDisplay *);
extern void       RXFlushWindow (RXDisplay *, Window);

extern void       RXUpdateMode (RXDisplay *, Window, int mode);

extern Font          RXLoadFont (RXDisplay *, const char *name);
extern RXFontStruct *RXQueryFont (RXDisplay *, Font);
extern RXFontStruct *RXLoadQueryFont (RXDisplay *, char *name);
extern void          RXFreeFont (RXDisplay *, RXFontStruct *);
extern void          RXUnloadFont (RXDisplay *, Font);

#endif
