/*
 * This file is part of Remote X11 (RX11), a library and server that lets
 * you develop applications on systems without X11 support.
 *
 * Copyright (C) 1995, Kay Roemer. See COPYING for details.
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "rx11.h"

#include "byteorder.h"

#define ALIGN4(x)   (((long)(x) + 3) & ~3L)

#define ROOM(r,n)   ((r)->ptr + ALIGN4 (n) <= &(r)->req[(r)->length])
#define SETPTR(r,n) ((r)->ptr += ALIGN4(n))


int
DecodeLong (Request *r, long *l)
{
    if (!ROOM (r, sizeof (*l)))
      return -1;

    *l = ntohl (*(long *)r->ptr);
    SETPTR (r, sizeof (*l));
    return 0;
}

int
DecodeShort (Request *r, short *s)
{
    if (!ROOM (r, sizeof (*s)))
      return -1;

    *s = ntohs (*(short *)r->ptr);
    SETPTR (r, sizeof (*s));
    return 0;
}

int
DecodeString (Request *r, char **s)
{
    long l;

    if (DecodeLong (r, &l))
      return -1;

    if (!ROOM (r, l+1))
      return -1;

    *s = r->ptr;
    SETPTR (r, l+1);
    return 0;
}

int
DecodeOpaque (Request *r, void **m, long *l)
{
    if (DecodeLong (r, l))
      return -1;

    if (!ROOM (r, *l))
      return -1;

    *m = r->ptr;
    SETPTR (r, *l);
    return 0;
}

int
DecodeRXVisual (Request *m, RXVisual *v)
{
  return
    DecodeLong  (m, &v->visualid) ||
    DecodeShort (m, &v->class) ||
    DecodeLong  (m, &v->red_mask) ||
    DecodeLong  (m, &v->green_mask) ||
    DecodeLong  (m, &v->green_mask) ||
    DecodeLong  (m, &v->blue_mask) ||
    DecodeShort (m, &v->bits_per_rgb) ||
    DecodeShort (m, &v->map_entries);
}

int
DecodeRXScreen (Request *m, RXScreen *s)
{
  return
    DecodeLong  (m, &s->root) ||
    DecodeShort (m, &s->width) ||
    DecodeShort (m, &s->height) ||
    DecodeShort (m, &s->mwidth) ||
    DecodeShort (m, &s->mheight) ||
    DecodeShort (m, &s->root_depth) ||
    DecodeLong  (m, &s->default_gc) ||
    DecodeLong  (m, &s->cmap) ||
    DecodeLong  (m, &s->white_pixel) ||
    DecodeLong  (m, &s->black_pixel) ||
    DecodeRXVisual (m, s->root_visual);
}

int
DecodeRXDisplay (Request *m, RXDisplay *d)
{
  return DecodeRXScreen (m, d->screen);
}

int
DecodeRXGCValues (Request *m, RXGCValues *v)
{
  register u_long mask;

  if (DecodeLong (m, &v->mask))
    return 1;

  mask = v->mask;

  return
    (mask & GCFunction        && DecodeShort (m, &v->function)) ||
    (mask & GCPlaneMask       && DecodeLong  (m, &v->plane_mask)) ||
    (mask & GCForeground      && DecodeLong  (m, &v->foreground)) ||
    (mask & GCBackground      && DecodeLong  (m, &v->background)) ||
    (mask & GCLineWidth       && DecodeShort (m, &v->line_width)) ||
    (mask & GCLineStyle       && DecodeShort (m, &v->line_style)) ||
    (mask & GCCapStyle        && DecodeShort (m, &v->cap_style)) ||
    (mask & GCJoinStyle       && DecodeShort (m, &v->join_style)) ||
    (mask & GCFillStyle       && DecodeShort (m, &v->fill_style)) ||
    (mask & GCArcMode         && DecodeShort (m, &v->arc_mode)) ||
    (mask & GCTile            && DecodeLong  (m, &v->tile)) ||
    (mask & GCStipple         && DecodeLong  (m, &v->stipple)) ||
    (mask & GCTileStipXOrigin && DecodeShort (m, &v->ts_x_origin)) ||
    (mask & GCTileStipYOrigin && DecodeShort (m, &v->ts_y_origin)) ||
    (mask & GCFont            && DecodeLong  (m, &v->font)) ||
    (mask & GCSubwindowMode   && DecodeShort (m, &v->subwindow_mode)) ||
    (mask & GCClipXOrigin     && DecodeShort (m, &v->clip_x_origin)) ||
    (mask & GCClipYOrigin     && DecodeShort (m, &v->clip_y_origin)) ||
    (mask & GCClipMask        && DecodeLong  (m, &v->clip_mask)) ||
    (mask & GCDashOffset      && DecodeShort (m, &v->dash_offset)) ||
    (mask & GCDashList        && DecodeShort (m, &v->dashes));
}

int
DecodeRXSetWindowAttributes (Request *m, RXSetWindowAttributes *a)
{
  register u_long mask;

  if (DecodeLong (m, &a->mask))
    return 1;

  mask = a->mask;

  return
    (mask & CWBackPixmap   && DecodeLong (m, &a->background_pixmap)) ||
    (mask & CWBackPixel    && DecodeLong (m, &a->background_pixel)) ||
    (mask & CWBorderPixmap && DecodeLong (m, &a->border_pixmap)) ||
    (mask & CWBorderPixel  && DecodeLong (m, &a->border_pixel)) ||
    (mask & CWColormap     && DecodeLong (m, &a->colormap)) ||
    (mask & CWCursor       && DecodeLong (m, &a->cursor));
}

int
DecodeRXWindowAttributes (Request *m, RXWindowAttributes *a, VisualID *visual)
{
  return
    DecodeShort (m, &a->x) ||
    DecodeShort (m, &a->y) ||
    DecodeShort (m, &a->width) ||
    DecodeShort (m, &a->height) ||
    DecodeShort (m, &a->border_width) ||
    DecodeShort (m, &a->depth) ||
    DecodeLong  (m, visual) ||
    DecodeShort (m, &a->class) ||
    DecodeLong  (m, &a->colormap) ||
    DecodeShort (m, &a->colormap_installed) ||
    DecodeShort (m, &a->map_state);
}

int
DecodeRXWindowChanges (Request *m, RXWindowChanges *c)
{
  register u_long mask;

  if (DecodeLong (m, &c->mask))
    return 1;

  mask = c->mask;

  return
    (mask & CWX           && DecodeShort (m, &c->x)) ||
    (mask & CWY           && DecodeShort (m, &c->y)) ||
    (mask & CWWidth       && DecodeShort (m, &c->width)) ||
    (mask & CWHeight      && DecodeShort (m, &c->height)) ||
    (mask & CWBorderWidth && DecodeShort (m, &c->border_width)) ||
    (mask & CWSibling     && DecodeLong  (m, &c->sibling)) ||
    (mask & CWStackMode   && DecodeShort (m, &c->stack_mode));
}

int
DecodeRXColor (Request *m, RXColor *c)
{
  return
    DecodeLong  (m, &c->pixel) ||
    DecodeShort (m, &c->red) ||
    DecodeShort (m, &c->green) ||
    DecodeShort (m, &c->blue) ||
    DecodeShort (m, &c->flags);
}

int
DecodeRXSegment (Request *m, RXSegment *s)
{
  return
    DecodeShort (m, &s->x1) ||
    DecodeShort (m, &s->y1) ||
    DecodeShort (m, &s->x2) ||
    DecodeShort (m, &s->y2);
}

int
DecodeRXPoint (Request *m, RXPoint *p)
{
  return
    DecodeShort (m, &p->x) ||
    DecodeShort (m, &p->y);
}

int
DecodeRXRectangle (Request *m, RXRectangle *r)
{
  return
    DecodeShort (m, &r->x) ||
    DecodeShort (m, &r->y) ||
    DecodeShort (m, &r->width) ||
    DecodeShort (m, &r->height);
}

int
DecodeRXArc (Request *m, RXArc *a)
{
  return
    DecodeShort (m, &a->x) ||
    DecodeShort (m, &a->y) ||
    DecodeShort (m, &a->width) ||
    DecodeShort (m, &a->height) ||
    DecodeShort (m, &a->angle1) ||
    DecodeShort (m, &a->angle2);
}
