/*
 * This file is part of Remote X11 (RX11), a library and server that lets
 * you develop applications on systems without X11 support.
 *
 * Copyright (C) 1995, Kay Roemer. See COPYING for details.
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "rx11.h"
#include "RXlib.h"

#include "byteorder.h"

#define ALIGN4(x)   (((long)(x) + 3) & ~3L)

#define ROOM(r,n)   ((r)->ptr + ALIGN4 (n) <= &(r)->req[MAX_REQ_SIZE])
#define SETPTR(r,n) { (r)->length += ALIGN4 (n); (r)->ptr += ALIGN4(n); }


int
EncodeLong (Request *r, long l)
{
    if (!ROOM (r, sizeof (l)))
      return -1;

    *(long *)r->ptr = htonl (l);
    SETPTR (r, sizeof (l));
    return 0;
}

int
EncodeShort (Request *r, short s)
{
    if (!ROOM (r, sizeof (s)))
      return -1;

    *(short *)r->ptr = htons (s);
    SETPTR (r, sizeof (s));
    return 0;
}

int
EncodeString (Request *r, char *s)
{
    long l = strlen (s);

    if (!ROOM (r, sizeof (l) + l+1))
      return -1;

    if (EncodeLong (r, l))
      return -1;

    memcpy (r->ptr, s, l+1);
    SETPTR (r, l+1);
    return 0;
}

int
EncodeOpaque (Request *r, void *m, long l)
{
    if (!ROOM (r, sizeof (l) + l))
      return -1;

    if (EncodeLong (r, l))
      return -1;

    memcpy (r->ptr, m, l);
    SETPTR (r, l);
    return 0;
}

int
EncodeRXVisual (Request *m, RXVisual *v)
{
  return
    EncodeLong  (m, v->visualid) ||
    EncodeShort (m, v->class) ||
    EncodeLong  (m, v->red_mask) ||
    EncodeLong  (m, v->green_mask) ||
    EncodeLong  (m, v->green_mask) ||
    EncodeLong  (m, v->blue_mask) ||
    EncodeShort (m, v->bits_per_rgb) ||
    EncodeShort (m, v->map_entries);
}

int
EncodeRXScreen (Request *m, RXScreen *s)
{
  return
    EncodeLong  (m, s->root) ||
    EncodeShort (m, s->width) ||
    EncodeShort (m, s->height) ||
    EncodeShort (m, s->mwidth) ||
    EncodeShort (m, s->mheight) ||
    EncodeShort (m, s->root_depth) ||
    EncodeLong  (m, s->default_gc) ||
    EncodeLong  (m, s->cmap) ||
    EncodeLong  (m, s->white_pixel) ||
    EncodeLong  (m, s->black_pixel) ||
    EncodeRXVisual (m, s->root_visual);
}

int
EncodeRXDisplay (Request *m, RXDisplay *d)
{
  return EncodeRXScreen (m, d->screen);
}

int
EncodeRXGCValues (Request *m, RXGCValues *v)
{
  register u_long mask = v->mask;

  return
    EncodeLong  (m, v->mask) ||
    (mask & GCFunction        && EncodeShort (m, v->function)) ||
    (mask & GCPlaneMask       && EncodeLong  (m, v->plane_mask)) ||
    (mask & GCForeground      && EncodeLong  (m, v->foreground)) ||
    (mask & GCBackground      && EncodeLong  (m, v->background)) ||
    (mask & GCLineWidth       && EncodeShort (m, v->line_width)) ||
    (mask & GCLineStyle       && EncodeShort (m, v->line_style)) ||
    (mask & GCCapStyle        && EncodeShort (m, v->cap_style)) ||
    (mask & GCJoinStyle       && EncodeShort (m, v->join_style)) ||
    (mask & GCFillStyle       && EncodeShort (m, v->fill_style)) ||
    (mask & GCArcMode         && EncodeShort (m, v->arc_mode)) ||
    (mask & GCTile            && EncodeLong  (m, v->tile)) ||
    (mask & GCStipple         && EncodeLong  (m, v->stipple)) ||
    (mask & GCTileStipXOrigin && EncodeShort (m, v->ts_x_origin)) ||
    (mask & GCTileStipYOrigin && EncodeShort (m, v->ts_y_origin)) ||
    (mask & GCFont            && EncodeLong  (m, v->font)) ||
    (mask & GCSubwindowMode   && EncodeShort (m, v->subwindow_mode)) ||
    (mask & GCClipXOrigin     && EncodeShort (m, v->clip_x_origin)) ||
    (mask & GCClipYOrigin     && EncodeShort (m, v->clip_y_origin)) ||
    (mask & GCClipMask        && EncodeLong  (m, v->clip_mask)) ||
    (mask & GCDashOffset      && EncodeShort (m, v->dash_offset)) ||
    (mask & GCDashList        && EncodeShort (m, v->dashes));
}

int
EncodeRXSetWindowAttributes (Request *m, RXSetWindowAttributes *a)
{
  register u_long mask = a->mask;

  return
    EncodeLong (m, a->mask) ||
    (mask & CWBackPixmap   && EncodeLong (m, a->background_pixmap)) ||
    (mask & CWBackPixel    && EncodeLong (m, a->background_pixel)) ||
    (mask & CWBorderPixmap && EncodeLong (m, a->border_pixmap)) ||
    (mask & CWBorderPixel  && EncodeLong (m, a->border_pixel)) ||
    (mask & CWColormap     && EncodeLong (m, a->colormap)) ||
    (mask & CWCursor       && EncodeLong (m, a->cursor));
}

int
EncodeRXWindowAttributes (Request *m, RXWindowAttributes *a)
{
  return
    EncodeShort (m, a->x) ||
    EncodeShort (m, a->y) ||
    EncodeShort (m, a->width) ||
    EncodeShort (m, a->height) ||
    EncodeShort (m, a->border_width) ||
    EncodeShort (m, a->depth) ||
    EncodeLong  (m, a->visual->visualid) ||
    EncodeShort (m, a->class) ||
    EncodeLong  (m, a->colormap) ||
    EncodeShort (m, a->colormap_installed) ||
    EncodeShort (m, a->map_state);
}

int
EncodeRXWindowChanges (Request *m, RXWindowChanges *c)
{
  register u_long mask = c->mask;

  return
    EncodeLong (m, mask) ||
    (mask & CWX           && EncodeShort (m, c->x)) ||
    (mask & CWY           && EncodeShort (m, c->y)) ||
    (mask & CWWidth       && EncodeShort (m, c->width)) ||
    (mask & CWHeight      && EncodeShort (m, c->height)) ||
    (mask & CWBorderWidth && EncodeShort (m, c->border_width)) ||
    (mask & CWSibling     && EncodeLong  (m, c->sibling)) ||
    (mask & CWStackMode   && EncodeShort (m, c->stack_mode));
}

int
EncodeRXColor (Request *m, RXColor *c)
{
  return
    EncodeLong  (m, c->pixel) ||
    EncodeShort (m, c->red) ||
    EncodeShort (m, c->green) ||
    EncodeShort (m, c->blue) ||
    EncodeShort (m, c->flags);
}

int
EncodeRXSegment (Request *m, RXSegment *s)
{
  return
    EncodeShort (m, s->x1) ||
    EncodeShort (m, s->y1) ||
    EncodeShort (m, s->x2) ||
    EncodeShort (m, s->y2);
}

int
EncodeRXPoint (Request *m, RXPoint *p)
{
  return
    EncodeShort (m, p->x) ||
    EncodeShort (m, p->y);
}

int
EncodeRXRectangle (Request *m, RXRectangle *r)
{
  return
    EncodeShort (m, r->x) ||
    EncodeShort (m, r->y) ||
    EncodeShort (m, r->width) ||
    EncodeShort (m, r->height);
}

int
EncodeRXArc (Request *m, RXArc *a)
{
  return
    EncodeShort (m, a->x) ||
    EncodeShort (m, a->y) ||
    EncodeShort (m, a->width) ||
    EncodeShort (m, a->height) ||
    EncodeShort (m, a->angle1) ||
    EncodeShort (m, a->angle2);
}
