/*
 * This file is part of Remote X11 (RX11), a library and server that lets
 * you develop applications on systems without X11 support.
 *
 * Copyright (C) 1995, Kay Roemer. See COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "rx11.h"
#include "util.h"

static Request req;

static int
OpenConn (RXDisplay *dpy, char *dname)
{
    char *host, *display, *screen;
    struct hostent *hent;
    int s, ret;
    struct sockaddr_in sin;

    host = strtok (dname, ":");
    display = strtok (NULL, ":");
    screen = strtok (NULL, ".");

    if (!host || !*host) {
	fprintf (stderr, "RX11: no host name specified\n");
	return -1;
    }

    hent = gethostbyname (host);
    if (!hent) {
	fprintf (stderr, "cannot look up host %s\n", host);
	return -1;
    }
    dpy->server_addr    = *(u_long *)hent->h_addr;
    dpy->server_display = display ? atoi (display) : 0;
    dpy->server_screen  = screen ? atoi (screen) : 0;

    s = socket (PF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	perror ("RX11: socket");
	return -1;
    }

    sin.sin_family = AF_INET;
    sin.sin_port = htons (RX11_PORT);
    sin.sin_addr.s_addr = dpy->server_addr;
    
    ret = connect (s, (struct sockaddr *)&sin, sizeof (sin));
    if (ret < 0) {
	perror ("RX11: connect");
	return -1;
    }

    dpy->fd = s;
    return 0;
}

RXDisplay *
RXOpenDisplay (char *dname)
{
    char display[200];
    RXDisplay *dpy;

    if (!dname)
      dname = getenv ("DISPLAY");

    if (!dname) {
	fprintf (stderr, "RX11: no DISPLAY defined.\n");
	return NULL;
    }
    strcpy (display, dname);

    dpy = xmalloc (sizeof (RXDisplay));
    dpy->screen = xmalloc (sizeof (RXScreen));
    dpy->screen->root_visual = xmalloc (sizeof (RXVisual));

    if (OpenConn (dpy, display))
      goto bad;

    RequestAlloc (&req, OpenDisplayReq);
    if (EncodeShort (&req, dpy->server_display) ||
	EncodeShort (&req, dpy->server_screen) ||
	RequestSend (dpy->fd, &req) ||
	RequestRecv (dpy->fd, &req))
      goto bad;

    if (req.opcode != OpenDisplayRep ||
	DecodeRXDisplay (&req, dpy)) {
      fprintf (stderr, "RX11: cannot decode reply\n");
      goto bad;
    }
    return dpy;

 bad:
    free (dpy->screen->root_visual);
    free (dpy->screen);
    free (dpy);
    return NULL;
}

int
RXCloseDisplay (RXDisplay *dpy)
{
    RequestAlloc (&req, CloseDisplayReq);
    RequestSend (dpy->fd, &req);

    close (dpy->fd);
    free (dpy->screen->root_visual);
    free (dpy->screen);
    free (dpy);
    return 0;
}

Window
RXCreateSimpleWindow (RXDisplay *dpy, Window parent,
		      int x, int y, u_int width, u_int height, u_int border_w,
		      u_long border, u_long background)
{
  Window win;

  RequestAlloc (&req, CreateSimpleWindowReq);
  if (EncodeLong (&req, parent) ||
      EncodeLong (&req, x) ||
      EncodeLong (&req, y) ||
      EncodeLong (&req, width) ||
      EncodeLong (&req, height) ||
      EncodeLong (&req, border_w) ||
      EncodeLong (&req, border) ||
      EncodeLong (&req, background) ||
      RequestSend (dpy->fd, &req) ||
      RequestRecv (dpy->fd, &req)) {
    fprintf (stderr, "RX11: cannot send req\n");
    return 0;
  }
  if (req.opcode != CreateSimpleWindowRep ||
      DecodeLong (&req, &win)) {
    fprintf (stderr, "RX11: cannot decode reply\n");
    return 0;
  }
  return win;
}

void
RXDestroyWindow (RXDisplay *dpy, Window w)
{
  RequestAlloc (&req, DestroyWindowReq);
  if (EncodeLong (&req, w) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXStoreName (RXDisplay *dpy, Window w, char *name)
{
  RequestAlloc (&req, StoreNameReq);
  if (EncodeLong (&req, w) ||
      EncodeString (&req, name) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXDrawPoint (RXDisplay *dpy, Drawable d, RXGC gc, int x, int y)
{
  RequestAlloc (&req, DrawPointReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, x) ||
      EncodeLong (&req, y) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXDrawLine (RXDisplay *dpy, Drawable d, RXGC gc,
	    int x1, int y1, int x2, int y2)
{
  RequestAlloc (&req, DrawLineReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, x1) ||
      EncodeLong (&req, y1) ||
      EncodeLong (&req, x2) ||
      EncodeLong (&req, y2) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXDrawRectangle (RXDisplay *dpy, Drawable d, RXGC gc, int x, int y,
		 u_int w, u_int h)
{
  RequestAlloc (&req, DrawRectangleReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, x) ||
      EncodeLong (&req, y) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXDrawArc (RXDisplay *dpy, Drawable d, RXGC gc, int x, int y,
	   u_int w, u_int h, short a1, short a2)
{
  RequestAlloc (&req, DrawArcReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, x) ||
      EncodeLong (&req, y) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      EncodeShort (&req, a1) ||
      EncodeShort (&req, a2) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXFillRectangle (RXDisplay *dpy, Drawable d, RXGC gc, int x, int y,
		 u_int w, u_int h)
{
  RequestAlloc (&req, FillRectangleReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, x) ||
      EncodeLong (&req, y) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXFillArc (RXDisplay *dpy, Drawable d, RXGC gc, int x, int y,
	   u_int w, u_int h, short a1, short a2)
{
  RequestAlloc (&req, FillArcReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, x) ||
      EncodeLong (&req, y) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      EncodeShort (&req, a1) ||
      EncodeShort (&req, a2) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXDrawString (RXDisplay *dpy, Drawable d, RXGC gc, int x, int y,
	      char *data, int len)
{
  RequestAlloc (&req, DrawStringReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, x) ||
      EncodeLong (&req, y) ||
      EncodeOpaque (&req, data, len) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXDrawImageString (RXDisplay *dpy, Drawable d, RXGC gc, int x, int y,
		   char *data, int len)
{
  RequestAlloc (&req, DrawImageStringReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, x) ||
      EncodeLong (&req, y) ||
      EncodeOpaque (&req, data, len) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

RXGC
RXCreateGC (RXDisplay *dpy, Drawable d, u_long mask, RXGCValues *val)
{
  RXGC gc;

  RequestAlloc (&req, CreateGCReq);
  val->mask = mask;
  if (EncodeLong (&req, d) ||
      EncodeRXGCValues (&req, val) ||
      RequestSend (dpy->fd, &req) ||
      RequestRecv (dpy->fd, &req)) {
    fprintf (stderr, "RX11: cannot send req\n");
    return 0;
  }
  if (req.opcode != CreateGCRep ||
      DecodeLong (&req, &gc)) {
    fprintf (stderr, "RX11: cannot decode reply\n");
    return 0;
  }
  return gc;
}

void
RXChangeGC (RXDisplay *dpy, RXGC gc, u_long mask, RXGCValues *val)
{
  val->mask = mask;
  RequestAlloc (&req, ChangeGCReq);
  if (EncodeLong (&req, gc) ||
      EncodeRXGCValues (&req, val) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXFreeGC (RXDisplay *dpy, RXGC gc)
{
  RequestAlloc (&req, FreeGCReq);
  if (EncodeLong (&req, gc) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

Pixmap
RXCreatePixmap (RXDisplay *dpy, Drawable d, int w, int h, int depth)
{
  Pixmap pix;

  RequestAlloc (&req, CreatePixmapReq);
  if (EncodeLong (&req, d) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      EncodeLong (&req, depth) ||
      RequestSend (dpy->fd, &req) ||
      RequestRecv (dpy->fd, &req)) {
    fprintf (stderr, "RX11: cannot send req\n");
    return None;
  }
  if (req.opcode != CreatePixmapRep ||
      DecodeLong (&req, &pix)) {
    fprintf (stderr, "RX11: cannot decode reply\n");
    return None;
  }
  return pix;
}

Pixmap
RXCreateBitmapFromData (RXDisplay *dpy, Drawable d, char *data, int w, int h)
{
  Pixmap pix;

  RequestAlloc (&req, CreateBitmapFromDataReq);
  if (EncodeLong (&req, d) ||
      EncodeOpaque (&req, data, ((w+7)/8) * h) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      RequestSend (dpy->fd, &req) ||
      RequestRecv (dpy->fd, &req)) {
    fprintf (stderr, "RX11: cannot send req\n");
    return None;
  }
  if (req.opcode != CreatePixmapRep ||
      DecodeLong (&req, &pix)) {
    fprintf (stderr, "RX11: cannot decode reply\n");
    return None;
  }
  return pix;
}

Pixmap
RXCreatePixmapFromBitmapData (RXDisplay *dpy, Drawable d, char *data,
			      int w, int h, u_long fg, u_long bg,
			      int depth)
{
  Pixmap pix;

  RequestAlloc (&req, CreatePixmapFromBitmapDataReq);
  if (EncodeLong (&req, d) ||
      EncodeOpaque (&req, data, ((w+7)/8) * h) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      EncodeLong (&req, fg) ||
      EncodeLong (&req, bg) ||
      EncodeLong (&req, depth) ||
      RequestSend (dpy->fd, &req) ||
      RequestRecv (dpy->fd, &req)) {
    fprintf (stderr, "RX11: cannot send req\n");
    return None;
  }
  if (req.opcode != CreatePixmapRep ||
      DecodeLong (&req, &pix)) {
    fprintf (stderr, "RX11: cannot decode reply\n");
    return None;
  }
  return pix;
}

int
RXReadBitmapFile (RXDisplay *dpy, Drawable d, char *fname,
		 u_int *w, u_int *h, Pixmap *pix, int *xhot, int *yhot)
{
  char *data;

  data = RXReadXbmFile (dpy, fname, w, h, xhot, yhot);
  if (!data)
    return 1;

  *pix = RXCreateBitmapFromData (dpy, d, data, *w, *h);
  return 0;
}

void
RXFreePixmap (RXDisplay *dpy, Pixmap p)
{
  RequestAlloc (&req, FreePixmapReq);
  if (EncodeLong (&req, p) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXCopyBitmap (RXDisplay *dpy, Drawable d, RXGC gc,
	      char *data, u_int width, u_int height,
	      int sx, int sy, u_int w, u_int h,
	      int dx, int dy)
{
  RequestAlloc (&req, CopyBitmapReq);

  if (EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeOpaque (&req, data, ((w+7)/8) * h) ||
      EncodeLong (&req, width) ||
      EncodeLong (&req, height) ||
      EncodeLong (&req, sx) ||
      EncodeLong (&req, sy) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      EncodeLong (&req, dx) ||
      EncodeLong (&req, dy) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXCopyArea (RXDisplay *dpy, Drawable s, Drawable d, RXGC gc,
	    int sx, int sy, u_int w, u_int h, int dx, int dy)
{
  RequestAlloc (&req, CopyAreaReq);
  if (EncodeLong (&req, s) ||
      EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, sx) ||
      EncodeLong (&req, sy) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      EncodeLong (&req, dx) ||
      EncodeLong (&req, dy) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXCopyPlane (RXDisplay *dpy, Drawable s, Drawable d, RXGC gc,
	     int sx, int sy, u_int w, u_int h, int dx, int dy,
	     u_long plane)
{
  RequestAlloc (&req, CopyPlaneReq);
  if (EncodeLong (&req, s) ||
      EncodeLong (&req, d) ||
      EncodeLong (&req, gc) ||
      EncodeLong (&req, sx) ||
      EncodeLong (&req, sy) ||
      EncodeLong (&req, w) ||
      EncodeLong (&req, h) ||
      EncodeLong (&req, dx) ||
      EncodeLong (&req, dy) ||
      EncodeLong (&req, plane) ||
      RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

int
RXAllocColor (RXDisplay *dpy, Colormap cmap, RXColor *col)
{
  long l;

  RequestAlloc (&req, AllocColorReq);
  if (EncodeLong (&req, cmap) ||
      EncodeRXColor (&req, col) ||
      RequestSend (dpy->fd, &req) ||
      RequestRecv (dpy->fd, &req)) {
    fprintf (stderr, "RX11: cannot send req\n");
    return 0;
  }
  switch (req.opcode) {
  case AllocColorRep:
    if (!DecodeRXColor (&req, col))
      return 1;

  case ErrorRep:
    if (!DecodeLong (&req, &l) && l == 1)
      return 0;
    break;
  }
  fprintf (stderr, "RX11: cannot decode reply\n");
  return 0;
}

int
RXAllocNamedColor (RXDisplay *dpy, Colormap cmap, char *name,
		   RXColor *scr_def, RXColor *exact_def)
{
  long l;

  RequestAlloc (&req, AllocNamedColorReq);
  if (EncodeLong (&req, cmap) ||
      EncodeString (&req, name) ||
      RequestSend (dpy->fd, &req) ||
      RequestRecv (dpy->fd, &req)) {
    fprintf (stderr, "RX11: cannot send req\n");
    return 0;
  }
  switch (req.opcode) {
  case AllocNamedColorRep:
    if (!DecodeRXColor (&req, scr_def) &&
	!DecodeRXColor (&req, exact_def))
      return 1;

  case ErrorRep:
    if (!DecodeLong (&req, &l) && l == 1)
      return 0;
    break;
  }
  fprintf (stderr, "RX11: cannot decode reply\n");
  return 0;
}

int
RXQueryColor (RXDisplay *dpy, Colormap cmap, RXColor *col)
{
  long l;

  RequestAlloc (&req, QueryColorReq);
  if (EncodeLong (&req, cmap) ||
      EncodeRXColor (&req, col) ||
      RequestSend (dpy->fd, &req) ||
      RequestRecv (dpy->fd, &req)) {
    fprintf (stderr, "RX11: cannot send req\n");
    return 0;
  }
  switch (req.opcode) {
  case QueryColorRep:
    if (!DecodeRXColor (&req, col))
      return 1;

  case ErrorRep:
    if (!DecodeLong (&req, &l) && l == 1)
      return 0;
    break;
  }
  fprintf (stderr, "RX11: cannot decode reply\n");
  return 0;
}

void
RXFlush (RXDisplay *dpy)
{
  RequestAlloc (&req, FlushReq);
  if (RequestSend (dpy->fd, &req))
    fprintf (stderr, "RX11: cannot send req\n");
}

void
RXFlushWindow (RXDisplay *dpy, Window w)
{
    RequestAlloc (&req, FlushWindowReq);
    if (EncodeLong (&req, w) ||
	RequestSend (dpy->fd, &req))
      fprintf (stderr, "RX11: cannot send req\n");
}

void
RXUpdateMode (RXDisplay *dpy, Window w, int mode)
{
    RequestAlloc (&req, UpdateModeReq);
    if (EncodeLong (&req, w) ||
	EncodeLong (&req, mode) ||
	RequestSend (dpy->fd, &req))
      fprintf (stderr, "RX11: cannot send req\n");
}

void
RXFree (void *p)
{
  if (p)
    free (p);
}
