/*
 * This file is part of Remote X11 (RX11), a library and server that lets
 * you develop applications on systems without X11 support.
 *
 * Copyright (C) 1995, Kay Roemer. See COPYING for details.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include "server.h"
#include "rx11.h"
#include "funcs.h"

Display *dpy;
short screen;
Visual *visual;

static void
ConvertXColor (RXColor *rcol, XColor *col, int dir)
{
  if (dir) {
    col->pixel = rcol->pixel;
    col->red = rcol->red;
    col->green = rcol->green;
    col->blue = rcol->blue;
    col->flags = rcol->flags;
  } else {
    rcol->pixel = col->pixel;
    rcol->red = col->red;
    rcol->green = col->green;
    rcol->blue = col->blue;
    rcol->flags = col->flags;
  }
}

static void
ConvertGCValues (RXGCValues *rval, XGCValues *val, u_long mask)
{
  val->function       = rval->function;
  val->plane_mask     = rval->plane_mask;
  val->foreground     = rval->foreground;
  val->background     = rval->background;
  val->line_width     = rval->line_width;
  val->line_style     = rval->line_style;
  val->cap_style      = rval->cap_style;
  val->join_style     = rval->join_style;
  val->fill_style     = rval->fill_style;
  val->arc_mode       = rval->arc_mode;
  if (mask & GCTile)
    val->tile         = WndLookWin (rval->tile);
  if (mask & GCStipple)
    val->stipple      = WndLookWin (rval->stipple);
  val->ts_x_origin    = rval->ts_x_origin;
  val->ts_y_origin    = rval->ts_y_origin;
  val->font           = rval->font;
  val->subwindow_mode = rval->subwindow_mode;
  val->clip_x_origin  = rval->clip_x_origin;
  val->clip_y_origin  = rval->clip_y_origin;
  if (mask & GCClipMask)
    val->clip_mask    = WndLookWin (rval->clip_mask);
  val->dash_offset    = rval->dash_offset;
  val->dashes         = rval->dashes;
}

/*
 *
 */

static void
SendErrorRep (long error, int fd)
{
  Request req;

  RequestAlloc (&req, ErrorRep);
  EncodeLong (&req, error);
  RequestSend (fd, &req);
}

void
DoOpenDisplay (Request *req, int fd)
{
  char *dname, new_dname[200], orig_dname[200];
  Screen  *scr;
  short display;
  RXDisplay rdpy;
  RXScreen  rscreen;
  RXVisual  rvisual;

  if (DecodeShort (req, &display) ||
      DecodeShort (req, &screen)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }

  dname = getenv ("DISPLAY");
  if (dname) {
    strcpy (orig_dname, dname);
    dname = strtok (orig_dname, ":");
  }
  sprintf (new_dname, "%s:%d.%d", dname ? dname : "", display, screen);

  dpy = XOpenDisplay (new_dname);
  if (!dpy) {
    fprintf (stderr, "cannot open display %s\n", new_dname);
    SendErrorRep (0, fd);
  }
  screen = DefaultScreen (dpy);
  scr    = DefaultScreenOfDisplay (dpy);
  visual = DefaultVisual (dpy, screen);

  rvisual.visualid     = visual->visualid;
  rvisual.class        = visual->class;
  rvisual.red_mask     = visual->red_mask;
  rvisual.green_mask   = visual->green_mask;
  rvisual.blue_mask    = visual->blue_mask;
  rvisual.bits_per_rgb = visual->bits_per_rgb;
  rvisual.map_entries  = visual->map_entries;

  rscreen.root        = WndAddSimple (RootWindowOfScreen (scr));
  rscreen.width       = WidthOfScreen (scr);
  rscreen.height      = HeightOfScreen (scr);
  rscreen.mwidth      = WidthMMOfScreen (scr);
  rscreen.mheight     = HeightMMOfScreen (scr);
  rscreen.root_depth  = PlanesOfScreen (scr);
  rscreen.root_visual = &rvisual;
  rscreen.default_gc  = GCAdd (DefaultGCOfScreen (scr));
  rscreen.cmap        = DefaultColormapOfScreen (scr);
  rscreen.white_pixel = WhitePixelOfScreen (scr);
  rscreen.black_pixel = BlackPixelOfScreen (scr);

  rdpy.screen = &rscreen;

  RequestAlloc (req, OpenDisplayRep);
  if (EncodeRXDisplay (req, &rdpy) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}

void
DoCloseDisplay (Request *req, int fd)
{
  if (dpy)
    XCloseDisplay (dpy);
}

void
DoFlush (Request *req, int fd)
{
    XFlush (dpy);
}

void
DoUpdateMode (Request *req, int fd)
{
    Window w;
    long mode;

    if (DecodeLong (req, &w) ||
	DecodeLong (req, &mode)) {
      fprintf (stderr, "cannot decode args\n");
      return;
    }
    WndSetUpdateMode (w, mode);
}

void
DoFlushWindow (Request *req, int fd)
{
    Window w;

    if (DecodeLong (req, &w)) {
	fprintf (stderr, "cannot decode args\n");
	return;
    }
    WndFlush (w);
}

void
DoCreateSimpleWindow (Request *req, int fd)
{
  Window parent, win;
  long x, y;
  u_long width, height, border_w;
  u_long border, bg;

  if (DecodeLong (req, &parent) ||
      DecodeLong (req, &x) ||
      DecodeLong (req, &y) ||
      DecodeLong (req, &width) ||
      DecodeLong (req, &height) ||
      DecodeLong (req, &border_w) ||
      DecodeLong (req, &border) ||
      DecodeLong (req, &bg)) {
    fprintf (stderr, "cannot decode args\n");
    SendErrorRep (0, fd);
    return;
  }
  win = XCreateSimpleWindow (dpy, WndLookWin (parent),
			     x, y, width, height, border_w,
			     border, bg);

  win = WndAdd (win, x, y, width, height);

  RequestAlloc (req, CreateSimpleWindowRep);
  if (EncodeLong (req, win) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}

void
DoDestroyWindow (Request *req, int fd)
{
  Window w;

  if (DecodeLong (req, &w)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XDestroyWindow (dpy, WndLookWin (w));
  WndDel (w);
}

void
DoStoreName (Request *req, int fd)
{
  Window w;
  char *name;

  if (DecodeLong (req, &w) ||
      DecodeString (req, &name)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XStoreName (dpy, WndLookWin (w), name);
}

void
DoDrawPoint (Request *req, int fd)
{
  Drawable d;
  RXGC gc;
  long x, y;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &x) ||
      DecodeLong (req, &y)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XDrawPoint (dpy, WndLookPix (d), GCLook (gc), x, y);
  if (WndNeedUpdate (d))
    XDrawPoint (dpy, WndLookWin (d), GCLook (gc), x, y);
}

void
DoDrawLine (Request *req, int fd)
{
  Drawable d;
  RXGC gc;
  long x1, y1, x2, y2;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &x1) ||
      DecodeLong (req, &y1) ||
      DecodeLong (req, &x2) ||
      DecodeLong (req, &y2)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XDrawLine (dpy, WndLookPix (d), GCLook (gc), x1, y1, x2, y2);
  if (WndNeedUpdate (d))
    XDrawLine (dpy, WndLookWin (d), GCLook (gc), x1, y1, x2, y2);
}

void
DoDrawRectangle (Request *req, int fd)
{
  Drawable d;
  RXGC gc;
  long x, y, w, h;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &x) ||
      DecodeLong (req, &y) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XDrawRectangle (dpy, WndLookPix (d), GCLook (gc), x, y, w, h);
  if (WndNeedUpdate (d))
    XDrawRectangle (dpy, WndLookWin (d), GCLook (gc), x, y, w, h);
}

void
DoFillRectangle (Request *req, int fd)
{
  Drawable d;
  RXGC gc;
  long x, y, w, h;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &x) ||
      DecodeLong (req, &y) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XFillRectangle (dpy, WndLookPix (d), GCLook (gc), x, y, w, h);
  WndUpdate (d, x, y, w, h);
}

void
DoDrawArc (Request *req, int fd)
{
  Drawable d;
  RXGC gc;
  long x, y, w, h;
  short a1, a2;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &x) ||
      DecodeLong (req, &y) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h) ||
      DecodeShort (req, &a1) ||
      DecodeShort (req, &a2)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XDrawArc (dpy, WndLookPix (d), GCLook (gc), x, y, w, h, a1, a2);
  if (WndNeedUpdate (d))
    XDrawArc (dpy, WndLookWin (d), GCLook (gc), x, y, w, h, a1, a2);
}

void
DoFillArc (Request *req, int fd)
{
  Drawable d;
  RXGC gc;
  long x, y, w, h;
  short a1, a2;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &x) ||
      DecodeLong (req, &y) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h) ||
      DecodeShort (req, &a1) ||
      DecodeShort (req, &a2)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XFillArc (dpy, WndLookPix (d), GCLook (gc), x, y, w, h, a1, a2);
  WndUpdate (d, x, y, w, h);
}

void
DoDrawString (Request *req, int fd)
{
  long x, y, len;
  void *data;
  RXGC gc;
  Drawable d;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &x) ||
      DecodeLong (req, &y) ||
      DecodeOpaque (req, &data, &len)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XDrawString (dpy, WndLookPix (d), GCLook (gc), x, y, data, len);
  if (WndNeedUpdate (d))
    XDrawString (dpy, WndLookWin (d), GCLook (gc), x, y, data, len);
}

void
DoDrawImageString (Request *req, int fd)
{
  long x, y, len;
  void *data;
  RXGC gc;
  Drawable d;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &x) ||
      DecodeLong (req, &y) ||
      DecodeOpaque (req, &data, &len)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XDrawImageString (dpy, WndLookPix (d), GCLook (gc), x, y, data, len);
  if (WndNeedUpdate (d))
    XDrawImageString (dpy, WndLookWin (d), GCLook (gc), x, y, data, len);
}

void
DoCreateGC (Request *req, int fd)
{
  Drawable d;
  RXGCValues rval;
  XGCValues val;
  GC gc;

  if (DecodeLong (req, &d) ||
      DecodeRXGCValues (req, &rval)) {
    fprintf (stderr, "cannot decode args\n");
    SendErrorRep (0, fd);
    return;
  }
  ConvertGCValues (&rval, &val, rval.mask);
  gc = XCreateGC (dpy, WndLookWin (d), rval.mask, &val);

  RequestAlloc (req, CreateGCRep);
  if (EncodeLong (req, GCAdd (gc)) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}

void
DoChangeGC (Request *req, int fd)
{
  RXGCValues rval;
  XGCValues val;
  RXGC gc;

  if (DecodeLong (req, &gc) ||
      DecodeRXGCValues (req, &rval)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  ConvertGCValues (&rval, &val, rval.mask);
  XChangeGC (dpy, GCLook (gc), rval.mask, &val);
}

void
DoFreeGC (Request *req, int fd)
{
  RXGC gc;

  if (DecodeLong (req, &gc)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XFreeGC (dpy, GCLook (gc));
  GCDel (gc);
}

void
DoCreatePixmap (Request *req, int fd)
{
  Drawable d;
  long w, h, depth;
  Pixmap pix;

  if (DecodeLong (req, &d) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h) ||
      DecodeLong (req, &depth)) {
    fprintf (stderr, "cannot decode args\n");
    SendErrorRep (0, fd);
    return;
  }
  pix = XCreatePixmap (dpy, WndLookWin (d), w, h, depth);
  if (pix)
    pix = WndAddSimple (pix);

  RequestAlloc (req, CreatePixmapRep);
  if (EncodeLong (req, pix) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}

void
DoCreateBitmapFromData (Request *req, int fd)
{
  Drawable d;
  long w, h, length;
  Pixmap pix;
  void *data;

  if (DecodeLong (req, &d) ||
      DecodeOpaque (req, &data, &length) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h)) {
    fprintf (stderr, "cannot decode args\n");
    SendErrorRep (0, fd);
    return;
  }
  pix = XCreateBitmapFromData (dpy, WndLookWin (d), data, w, h);
  if (pix)
    pix = WndAddSimple (pix);

  RequestAlloc (req, CreatePixmapRep);
  if (EncodeLong (req, pix) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}

void
DoCreatePixmapFromBitmapData (Request *req, int fd)
{
  Drawable d;
  long w, h, length;
  Pixmap pix;
  void *data;
  u_long fg, bg;
  u_long depth;

  if (DecodeLong (req, &d) ||
      DecodeOpaque (req, &data, &length) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h) ||
      DecodeLong (req, &fg) ||
      DecodeLong (req, &bg) ||
      DecodeLong (req, &depth)) {
    fprintf (stderr, "cannot decode args\n");
    SendErrorRep (0, fd);
    return;
  }
  pix = XCreatePixmapFromBitmapData (dpy, WndLookWin (d), data, w, h,
				     fg, bg, depth);
  if (pix)
    pix = WndAddSimple (pix);

  RequestAlloc (req, CreatePixmapRep);
  if (EncodeLong (req, pix) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}

void
DoFreePixmap (Request *req, int fd)
{
  Pixmap pix;

  if (DecodeLong (req, &pix)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XFreePixmap (dpy, WndLookWin (pix));
  WndDel (pix);
}

void
DoCopyArea (Request *req, int fd)
{
  Drawable s, d;
  RXGC gc;
  long sx, sy, dx, dy;
  u_long w, h;

  if (DecodeLong (req, &s) ||
      DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &sx) ||
      DecodeLong (req, &sy) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h) ||
      DecodeLong (req, &dx) ||
      DecodeLong (req, &dy)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XCopyArea (dpy, WndLookPix (s), WndLookPix (d), GCLook (gc),
	     sx, sy, w, h, dx, dy);
  WndUpdate (d, dx, dy, w, h);
}

void
DoCopyPlane (Request *req, int fd)
{
  Drawable s, d;
  RXGC gc;
  long sx, sy, dx, dy;
  u_long w, h, plane;

  if (DecodeLong (req, &s) ||
      DecodeLong (req, &d) ||
      DecodeLong (req, &gc) ||
      DecodeLong (req, &sx) ||
      DecodeLong (req, &sy) ||
      DecodeLong (req, &w) ||
      DecodeLong (req, &h) ||
      DecodeLong (req, &dx) ||
      DecodeLong (req, &dy) ||
      DecodeLong (req, &plane)) {
    fprintf (stderr, "cannot decode args\n");
    return;
  }
  XCopyPlane (dpy, WndLookPix (s), WndLookPix (d), GCLook (gc),
	      sx, sy, w, h, dx, dy, plane);
  WndUpdate (d, dx, dy, w, h);
}

void
DoCopyBitmap (Request *req, int fd)
{
    Pixmap pix;
    Drawable d;
    RXGC gc;
    void *data;
    long width, height, length;
    long sx, sy, w, h, dx, dy;

    if (DecodeLong (req, &d) ||
	DecodeLong (req, &gc) ||
	DecodeOpaque (req, &data, &length) ||
	DecodeLong (req, &width) ||
	DecodeLong (req, &height) ||
	DecodeLong (req, &sx) ||
	DecodeLong (req, &sy) ||
	DecodeLong (req, &w) ||
	DecodeLong (req, &h) ||
	DecodeLong (req, &dx) ||
	DecodeLong (req, &dy)) {
      fprintf (stderr, "cannot decode args\n");
      return;
    }
    pix = XCreateBitmapFromData (dpy, WndLookWin (d), data,
				 width, height);
    if (pix) {
	XCopyPlane (dpy, pix, WndLookPix (d), GCLook (gc),
		    sx, sy, w, h, dx, dy, 1);
	WndUpdate (d, dx, dy, w, h);
	XFreePixmap (dpy, pix);
    }
}

void
DoAllocColor (Request *req, int fd)
{
  Colormap cmap;
  RXColor rcol;
  XColor col;
  
  if (DecodeLong (req, &cmap) ||
      DecodeRXColor (req, &rcol)) {
    fprintf (stderr, "cannot decode args\n");
    SendErrorRep (0, fd);
    return;
  }
  ConvertXColor (&rcol, &col, 1);
  if (!XAllocColor (dpy, cmap, &col)) {
    SendErrorRep (1, fd);
    return;
  }
  ConvertXColor (&rcol, &col, 0);

  RequestAlloc (req, AllocColorRep);
  if (EncodeRXColor (req, &rcol) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}

void
DoAllocNamedColor (Request *req, int fd)
{
  Colormap cmap;
  RXColor rcol1, rcol2;
  XColor col1, col2;
  char *name;
  
  if (DecodeLong (req, &cmap) ||
      DecodeString (req, &name)) {
    fprintf (stderr, "cannot decode args\n");
    SendErrorRep (0, fd);
    return;
  }
  if (!XAllocNamedColor (dpy, cmap, name, &col1, &col2)) {
    SendErrorRep (1, fd);
    return;
  }
  ConvertXColor (&rcol1, &col1, 0);
  ConvertXColor (&rcol2, &col2, 0);

  RequestAlloc (req, AllocNamedColorRep);
  if (EncodeRXColor (req, &rcol1) ||
      EncodeRXColor (req, &rcol2) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}

void
DoQueryColor (Request *req, int fd)
{
  Colormap cmap;
  RXColor rcol;
  XColor col;
  
  if (DecodeLong (req, &cmap) ||
      DecodeRXColor (req, &rcol)) {
    fprintf (stderr, "cannot decode args\n");
    SendErrorRep (0, fd);
    return;
  }
  ConvertXColor (&rcol, &col, 1);
  if (!XQueryColor (dpy, cmap, &col)) {
    SendErrorRep (1, fd);
    return;
  }
  ConvertXColor (&rcol, &col, 0);

  RequestAlloc (req, QueryColorRep);
  if (EncodeRXColor (req, &rcol) ||
      RequestSend (fd, req))
    fprintf (stderr, "cannot send reply\n");
}
