/*
 * This file is part of Remote X11 (RX11), a library and server that lets
 * you develop applications on systems without X11 support.
 *
 * Copyright (C) 1995, Kay Roemer. See COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "server.h"
#include "rx11.h"
#include "funcs.h"
#include "version.h"

static inline long
timedif (struct timeval *this, struct timeval *last)
{
  return (long)(this->tv_sec - last->tv_sec) * 1000L +
         (long)(this->tv_usec - last->tv_usec) / 1000L;
}

/*
 * poll every 100 ms
 */
#define POLL 100

static int
DispatchRequest (Request *req, int nsock)
{
  switch (req->opcode) {
  case OpenDisplayReq:
    DoOpenDisplay (req, nsock);
    break;

  case CloseDisplayReq:
    DoCloseDisplay (req, nsock);
    return 1;

  case FlushReq:
    DoFlush (req, nsock);
    break;

  case FlushWindowReq:
    DoFlushWindow (req, nsock);
    break;

  case UpdateModeReq:
    DoUpdateMode (req, nsock);
    break;

  case CreateSimpleWindowReq:
    DoCreateSimpleWindow (req, nsock);
    break;

  case DestroyWindowReq:
    DoDestroyWindow (req, nsock);
    break;

  case StoreNameReq:
    DoStoreName (req, nsock);
    break;

  case DrawPointReq:
    DoDrawPoint (req, nsock);
    break;

  case DrawLineReq:
    DoDrawLine (req, nsock);
    break;

  case DrawRectangleReq:
    DoDrawRectangle (req, nsock);
    break;

  case DrawArcReq:
    DoDrawArc (req, nsock);
    break;

  case FillRectangleReq:
    DoFillRectangle (req, nsock);
    break;

  case FillArcReq:
    DoFillArc (req, nsock);
    break;

  case DrawStringReq:
    DoDrawString (req, nsock);
    break;

  case DrawImageStringReq:
    DoDrawImageString (req, nsock);
    break;

  case CreateGCReq:
    DoCreateGC (req, nsock);
    break;

  case ChangeGCReq:
    DoChangeGC (req, nsock);
    break;

  case FreeGCReq:
    DoFreeGC (req, nsock);
    break;

  case CreatePixmapReq:
    DoCreatePixmap (req, nsock);
    break;

  case CreateBitmapFromDataReq:
    DoCreateBitmapFromData (req, nsock);
    break;

  case CreatePixmapFromBitmapDataReq:
    DoCreatePixmapFromBitmapData (req, nsock);
    break;

  case FreePixmapReq:
    DoFreePixmap (req, nsock);
    break;

  case CopyAreaReq:
    DoCopyArea (req, nsock);
    break;

  case CopyPlaneReq:
    DoCopyPlane (req, nsock);
    break;

  case CopyBitmapReq:
    DoCopyBitmap (req, nsock);
    break;

  case AllocColorReq:
    DoAllocColor (req, nsock);
    break;

  case QueryColorReq:
    DoQueryColor (req, nsock);
    break;

  case AllocNamedColorReq:
    DoAllocNamedColor (req, nsock);
    break;

  default:
    fprintf (stderr, "opcode mismatch\n");
    break;
  }
  return 0;
}

static void
ServeClient (int nsock, struct sockaddr_in *sin)
{
    Request req;
    fd_set rfd;
    struct timeval tm, last, this;
    int r;

    printf ("new client from %s:%d\n", inet_ntoa (sin->sin_addr),
	    sin->sin_port);
    gettimeofday (&last, 0);

    while (42) {
        FD_ZERO (&rfd);
        FD_SET (nsock, &rfd);
	tm.tv_sec = 0;
	tm.tv_usec = POLL*1000L;

	r = select (nsock+1, &rfd, 0, 0, &tm);
	if (dpy) {
	  gettimeofday (&this, 0);
	  if (timedif (&this, &last) > POLL) {
	    last = this;
	    XFlush (dpy);
	    if (XPending (dpy))
	      WndDoEvents ();
	  }
	}
	if (r != 1)
	  continue;

	RequestAlloc (&req, OpenDisplayReq);
	r = RequestRecv (nsock, &req);
	if (r == 1) {
	  fprintf (stderr, "connection broken\n");
	  goto leave;
	} else if (r < 0) {
	    fprintf (stderr, "recv error\n");
	    goto leave;
	}
	if (DispatchRequest (&req, nsock))
	  goto leave;
    }

 leave:
    printf ("server exit\n");
    exit (0);
}

int
main (int argc, char *argv[])
{
    struct sockaddr_in sin;
    int ret, sock;

    printf ("Remote X11 server %s, Copyright (C) 1995 Kay Roemer.\n",
            RX11_VERSION);

    sock = socket (PF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	perror ("socket");
	return 0;
    }

    sin.sin_family = AF_INET;
    sin.sin_port = htons (RX11_PORT);
    sin.sin_addr.s_addr = INADDR_ANY;

    ret = bind (sock, (struct sockaddr *)&sin, sizeof (sin));
    if (ret < 0) {
	perror ("bind");
	return 0;
    }
    listen (sock, 5);
    while (42) {
	int pid, nsock, addrlen;

	addrlen = sizeof (sin);
	nsock = accept (sock, (struct sockaddr *)&sin, &addrlen);
	if (nsock >= 0) {
	    pid = fork ();
	    if (pid < 0)
	      perror ("fork");
	    else if (pid == 0)
	      ServeClient (nsock, &sin);
	    else
	      close (nsock);
	} else
	  perror ("accept");
    }
    return 0;
}


