/*
 * This file is part of Remote X11 (RX11), a library and server that lets
 * you develop applications on systems without X11 support.
 *
 * Copyright (C) 1995, Kay Roemer. See COPYING for details.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include "server.h"
#include "rx11.h"
#include "funcs.h"

#define MAX_WND 16

static struct WndTabEntry {
  int    x, y;
  u_int  width, height;
  Window win;
  Pixmap backing;
  u_int  backing_width, backing_height;
  int    depth;
  GC     gc;
  int    simple;
  int    mapped;
  int    used;
  int    dirty;
  int    updatemode;
} WndTab[MAX_WND];

static int Initialized = 0;

static void
Init (void)
{
  Initialized = 1;
  /*
   * Add mapping for None resource
   */
  WndAddSimple (None);
}

Window
WndAddSimple (Window win)
{
  int i;

  if (!Initialized)
    Init ();

  if ((u_long)win >= MAX_WND)
    return win;

  for (i = 0; i < MAX_WND; ++i) {
    if (!WndTab[i].used) {
      WndTab[i].used = 1;
      WndTab[i].simple = 1;
      WndTab[i].win = win;
      return i;
    }
  }
  fprintf (stderr, "window table overflow\n");
  return 0;
}

Window
WndAdd (Window win, int x, int y, u_int w, u_int h)
{
  XSizeHints sz_hints;
  XWindowAttributes w_attr;
  XGCValues gc_val;
  int i;

  if (!Initialized)
    Init ();

  for (i = 0; i < MAX_WND; ++i) {
    if (!WndTab[i].used)
      break;
  }
  if (i >= MAX_WND) {
    fprintf (stderr, "window table overflow\n");
    return 0;
  }
  XGetWindowAttributes (dpy, win, &w_attr);

  WndTab[i].mapped = 0;
  WndTab[i].simple = 0;
  WndTab[i].used = 1;
  WndTab[i].updatemode = 1;
  WndTab[i].dirty = 0;
  WndTab[i].x = x;
  WndTab[i].y = y;
  WndTab[i].width =
  WndTab[i].backing_width = w;
  WndTab[i].height =
  WndTab[i].backing_height = h;
  WndTab[i].depth = w_attr.depth;
  WndTab[i].win = win;
  WndTab[i].gc = XCreateGC (dpy, win, 0, &gc_val);
  WndTab[i].backing = XCreatePixmap (dpy, win, w, h, w_attr.depth);

  XSetForeground (dpy, WndTab[i].gc, WhitePixel (dpy, screen));
  XSetBackground (dpy, WndTab[i].gc, BlackPixel (dpy, screen));
  XFillRectangle (dpy, WndTab[i].backing, WndTab[i].gc, 0, 0, w, h);

  sz_hints.flags = PPosition | PSize;
  sz_hints.x = x;
  sz_hints.y = y;
  sz_hints.width = w;
  sz_hints.height = h;

  XSetWMNormalHints (dpy, win, &sz_hints);
  XStoreName (dpy, win, "RX11 Window");
  XSelectInput (dpy, win, ExposureMask | StructureNotifyMask);
  XMapWindow (dpy, win);
  XFlush (dpy);
  return i;
}

void
WndDel (Window rwin)
{
  if (rwin == None)
    return;
  if ((u_long)rwin < MAX_WND) {
    if (!WndTab[rwin].used)
      fprintf (stderr, "stale window handle\n");
    else {
      if (!WndTab[rwin].simple) {
	XFreeGC (dpy, WndTab[rwin].gc);
	XFreePixmap (dpy, WndTab[rwin].backing);
      }
      WndTab[rwin].used = 0;
    }
  }
}

void
WndSetUpdateMode (Window rwin, long mode)
{
  if ((u_long)rwin < MAX_WND) {
    if (!WndTab[rwin].used) {
      fprintf (stderr, "stale window handle\n");
      return;
    }
    if (!WndTab[rwin].updatemode && mode)
      WndFlush (rwin);
    WndTab[rwin].updatemode = mode;
  }
}

void
WndFlush (Window rwin)
{
    if ((u_long)rwin < MAX_WND) {
	if (!WndTab[rwin].used) {
	    fprintf (stderr, "stale window handle\n");
	    return;
	}
	if (WndTab[rwin].dirty && !WndTab[rwin].simple &&
	    WndTab[rwin].mapped) {
	    XCopyArea (dpy, WndTab[rwin].backing, WndTab[rwin].win,
		       WndTab[rwin].gc, 0, 0,
		       WndTab[rwin].width, WndTab[rwin].height, 0, 0);
	    WndTab[rwin].dirty = 0;
	}
    }
}

Drawable
WndLookPix (Window rwin)
{
  if ((u_long)rwin < MAX_WND) {
    if (!WndTab[rwin].used) {
      fprintf (stderr, "stale window handle\n");
      return None;
    }
    return WndTab[rwin].simple ? WndTab[rwin].win : WndTab[rwin].backing;
  }
  return rwin;
}

Window
WndLookWin (Window rwin)
{
  if ((u_long)rwin < MAX_WND) {
    if (!WndTab[rwin].used) {
      fprintf (stderr, "stale window handle\n");
      return None;
    }
    return WndTab[rwin].win;
  }
  return rwin;
}

void
WndUpdate (Window rwin, int x, int y, u_int w, u_int h)
{
  if ((u_long)rwin < MAX_WND) {
    if (!WndTab[rwin].used)
      fprintf (stderr, "stale window handle\n");
    else if (!WndTab[rwin].simple && WndTab[rwin].mapped) {
	if (WndTab[rwin].updatemode) {
	    if (w && h)
	      XCopyArea (dpy, WndTab[rwin].backing, WndTab[rwin].win,
			 WndTab[rwin].gc, x, y, w, h, x, y);
	    else
	      XCopyArea (dpy, WndTab[rwin].backing, WndTab[rwin].win,
			 WndTab[rwin].gc, 0, 0,
			 WndTab[rwin].width, WndTab[rwin].height, 0, 0);
	} else
	  WndTab[rwin].dirty = 1;
    }
  }
}

int
WndNeedUpdate (Window rwin)
{
  if ((u_long)rwin < MAX_WND) {
    if (!WndTab[rwin].used)
      fprintf (stderr, "stale window handle\n");
    else if (!WndTab[rwin].simple && WndTab[rwin].mapped) {
	if (WndTab[rwin].updatemode)
	  return 1;
	WndTab[rwin].dirty = 1;
    }
  }
  return 0;
}

static struct WndTabEntry *
WndLookRev (Window win)
{
  int i;

  for (i = 0; i < MAX_WND; ++i) {
    if (WndTab[i].used && WndTab[i].win == win)
      return &WndTab[i];
  }
  return 0;
}

void
WndDoEvents (void)
{
  struct WndTabEntry *wp;
  XEvent event;
  int nevents;

  for (nevents = QLength (dpy); --nevents >= 0; ) {
    XNextEvent (dpy, &event);

    switch (event.type) {
    case ConfigureNotify:
      wp = WndLookRev (event.xconfigure.window);
      if (wp && !wp->simple) {
	wp->width = event.xconfigure.width;
	wp->height = event.xconfigure.height;
	if (wp->width > wp->backing_width ||
	    wp->height > wp->backing_height) {
	  Pixmap nbacking;

	  nbacking = XCreatePixmap (dpy, wp->win,
				    wp->width, wp->height,
				    wp->depth);
	  XFillRectangle (dpy, nbacking, wp->gc, 0, 0, wp->width, wp->height);
	  XCopyArea (dpy, wp->backing, nbacking, wp->gc,
		     0, 0, wp->backing_width, wp->backing_height,
		     0, 0);
	  XFreePixmap (dpy, wp->backing);
	  wp->backing = nbacking;
	  wp->backing_width = wp->width;
	  wp->backing_height = wp->height;
	}
      }
      break;

    case Expose:
      wp = WndLookRev (event.xexpose.window);
      if (wp && !wp->simple) {
	wp->mapped = 1;
	XCopyArea (dpy, wp->backing, wp->win, wp->gc,
		   event.xexpose.x, event.xexpose.y,
		   event.xexpose.width, event.xexpose.height,
		   event.xexpose.x, event.xexpose.y);
      }
      break;
    }
  }
}
