/*
 * This file is part of Remote X11 (RX11), a library and server that lets
 * you develop applications on systems without X11 support.
 *
 * Copyright (C) 1995, Kay Roemer. See COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include "util.h"
#include "RXlib.h"
#include "rx11.h"

static int
remove_comment (FILE *fp)
{
  int c, lastc = 0;

  while ((c = fgetc (fp)) != EOF) {
    if (c == '*' && lastc == '/' && remove_comment (fp))
      return 1;
    if (c == '/' && lastc == '*')
      return 0;
    lastc = c;
  }
  return 1;
}

static int
skip (FILE *fp, int *cp)
{
  int c, lastc = 0;

  while ((c = fgetc (fp)) != EOF &&
	 (isspace (c) ||
	  c == ',' || c == ';' || c == '{' || c == '#' ||
	  c == '}' || c == ']' || c == '[' || c == '=' ||
	  c == '/' || c == '*')) {
    if (c == '*' && lastc == '/' && remove_comment (fp)) {
      *cp = EOF;
      return 1;
    }
    lastc = c;
  }
  *cp = c;
  if (c != EOF) {
    ungetc (c, fp);
    return 0;
  }
  return 1;
}

static int
number (FILE *fp, int *n)
{
  return fscanf (fp, "%i", n) != 1;
}

#define isstr1(c) (isalpha (c) || (c) == '_')
#define isstr2(c) (isalnum (c) || (c) == '_')

static int
string (FILE *fp, char **s)
{
  static char str[200];
  char *cp = str;
  int c;

  if ((c = fgetc (fp)) == EOF || !isstr1 (c)) {
    if (c != EOF)
      ungetc (c, fp);
    return 1;
  }
  *cp++ = c;
  while ((c = fgetc (fp)) != EOF && isstr2 (c))
    *cp++ = c;
  *cp = '\0';

  *s = str;
  return 0;
}

char *
RXReadXbmFile (RXDisplay *dpy, char *fname, u_int *width, u_int *height,
	       int *xhot, int *yhot)
{
  int w, h, len, i;
  char *data, *cp;
  int num, c;
  char *str;
  FILE *fp;

  data = 0;
  w = h = 0;
  *xhot = *yhot = 0;

  fp = fopen (fname, "r");
  if (!fp) {
    fprintf (stderr, "RX11: cannot open input file '%s'\n", fname);
    return 0;
  }
  for (i = 0;; ++i) {
    if (skip (fp, &c) || string (fp, &str) || strcmp (str, "define"))
      break;
    if (skip (fp, &c) || string (fp, &str) ||
	skip (fp, &c) || number (fp, &num)) {
      fprintf (stderr, "RX11: malformed xbm header\n");
      goto bad;
    }
    switch (i) {
    case 0:
      *width = w = num;
      break;
    case 1:
      *height = h = num;
      break;
    case 2:
      *xhot = num;
      break;
    case 3:
      *yhot = num;
      break;
    }
  }
  if (i < 2) {
    fprintf (stderr, "RX11: missing header fields in xbm file\n");
    goto bad;
  }
  while (!skip (fp, &c) && isstr1 (c) && !string (fp, &str))
    ;
  if (!isdigit (c)) {
    fprintf (stderr, "RX11: malformed xbm header\n");
    goto bad;
  }

  len = ((w+7)/8) * h;
  cp = data = xmalloc (len);
  for (i = 0; i < len; ++i) {
    if (skip (fp, &c) || number (fp, &num)) {
      fprintf (stderr, "RX11: malformed xbm data\n");
      goto bad;
    }
    *cp++ = (char)num;
  }
  fclose (fp);
  return data;

 bad:
  if (data)
    free (data);
  fclose (fp);
  return 0;
}

/*
 * Bitmap scaling routines
 */

static struct {
  u_char data[8];
} HMap[256];

static u_char Bits[8][8] = {
  { 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff },
  { 0x02, 0x06, 0x0e, 0x1e, 0x3e, 0x7e, 0xfe, 0xfe },
  { 0x04, 0x0c, 0x1c, 0x3c, 0x7c, 0xfc, 0xfc, 0xfc },
  { 0x08, 0x18, 0x38, 0x78, 0xf8, 0xf8, 0xf8, 0xf8 },
  { 0x10, 0x30, 0x70, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0 },
  { 0x20, 0x60, 0xe0, 0xe0, 0xe0, 0xe0, 0xe0, 0xe0 },
  { 0x40, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0 },
  { 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80 }
};

static struct {
  u_char op, offset;
} VMap[8];

static u_int  VSteps = 0;

static u_int LastHScale = 0;
static u_int LastVScale = 0;

static void
UpdateHMap (u_int hscale)
{
  int i, bit;
  u_char j, *cp;

  memset (HMap, 0, sizeof (HMap));
  for (i = 0; i < 256; ++i) {
    bit = 0;
    cp = HMap[i].data;
    for (j = 0x01; j; j <<= 1) {
      if (i & j) {
	*cp |= Bits[bit][hscale-1];
	if ((bit += hscale) >= 8) {
	  ++cp;
	  if ((bit -= 8))
	    *cp |= Bits[0][bit-1];
	}
      } else if ((bit += hscale) >= 8) {
	bit -= 8;
	++cp;
      }
    }
  }
  LastHScale = hscale;
}

static void
UpdateVMap (u_int vscale)
{
  int i;

  for (i = 0; vscale > 1; ++i) {
    if (vscale & 1) {
      VMap[i].op     = 0;
      VMap[i].offset = --vscale;
    } else {
      VMap[i].op     = 1;
      VMap[i].offset = (vscale >>= 1);
    }
  }
  VSteps = i;
  LastVScale = vscale;
}

char *
RXScaleBitmapData (RXDisplay *dpy, u_char *data, u_int *w, u_int *h,
		   u_int hscale, u_int vscale)
{
  int i, j, k;
  u_int nw, nh, bytes_per_line, nbytes_per_line;
  u_char *line_start, *cp, *cp2, *ndata;

  if (hscale < 1 || hscale > 8 || vscale < 1 || vscale > 8) {
    fprintf (stderr, "RX11: invalid scaling factors\n");
    return NULL;
  }

  bytes_per_line = (*w+7)/8;

  nw = *w * hscale;
  nh = *h * vscale;
  nbytes_per_line = (nw+7)/8;
  ndata = xmalloc (nbytes_per_line * nh + 8);

  if (hscale != LastHScale)
    UpdateHMap (hscale);

  if (vscale != LastVScale)
    UpdateVMap (vscale);

  line_start = ndata;
  for (k = *h; --k >= 0; ) {

    if (hscale != 1) {
      /*
       * stretch first line in X direction
       */
      cp = line_start;
      for (j = bytes_per_line; --j >= 0; ) {
	cp2 = HMap[*data++].data;
	for (i = hscale; --i >= 0; )
	  *cp++ = *cp2++;
      }
    } else {
      memcpy (line_start, data, nbytes_per_line);
      data += nbytes_per_line;
    }

    if (vscale != 1) {
      /*
       * duplicate newly generated line in Y direction
       */
      for (i = VSteps; --i >= 0; ) {
	u_int offset = nbytes_per_line * VMap[i].offset;

	memcpy (line_start + offset,
		line_start,
		VMap[i].op ? offset : nbytes_per_line);
      }
    }
    line_start += vscale * nbytes_per_line;
  }
  *w = nw;
  *h = nh;
  return ndata;
}
